/*
 * Michael MIC implementation - optimized for TKIP MIC operations
 * Copyright 2002-2003, Instant802 Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/types.h>
#include <linux/bitops.h>
#include <linux/ieee80211.h>
#include <asm/unaligned.h>

#include "michael.h"

static void michael_block(struct michael_mic_ctx *mctx, u32 val)
{
	mctx->l ^= val;
	mctx->r ^= rol32(mctx->l, 17);
	mctx->l += mctx->r;
	mctx->r ^= ((mctx->l & 0xff00ff00) >> 8) |
		   ((mctx->l & 0x00ff00ff) << 8);
	mctx->l += mctx->r;
	mctx->r ^= rol32(mctx->l, 3);
	mctx->l += mctx->r;
	mctx->r ^= ror32(mctx->l, 2);
	mctx->l += mctx->r;
}

static void michael_mic_hdr(struct michael_mic_ctx *mctx, const u8 *key,
			    struct ieee80211_hdr *hdr)
{
	u8 *da, *sa, tid;

	da = ieee80211_get_DA(hdr);
	sa = ieee80211_get_SA(hdr);
	if (ieee80211_is_data_qos(hdr->frame_control))
		tid = *ieee80211_get_qos_ctl(hdr) & IEEE80211_QOS_CTL_TID_MASK;
	else
		tid = 0;

	mctx->l = key[0] | key[1] << 8 | key[2] << 16 | key[3] << 24;
	mctx->r = key[4] | key[5] << 8 | key[6] << 16 | key[7] << 24;

	/*
	 * A pseudo header (DA, SA, Priority, 0, 0, 0) is used in Michael MIC
	 * calculation, but it is _not_ transmitted
	 */
	michael_block(mctx, da[0] | da[1] << 8 | da[2] << 16 | da[3] << 24);
	michael_block(mctx, da[4] | da[5] << 8 | sa[0] << 16 | sa[1] << 24);
	michael_block(mctx, sa[2] | sa[3] << 8 | sa[4] << 16 | sa[5] << 24);
	michael_block(mctx, tid);
}

void michael_mic(const u8 *key, struct ieee80211_hdr *hdr,
		 const u8 *data, size_t data_len, u8 *mic)
{
	u32 val;
	size_t block, blocks, left;
	struct michael_mic_ctx mctx;

	michael_mic_hdr(&mctx, key, hdr);

	/* Real data */
	blocks = data_len / 4;
	left = data_len % 4;

	for (block = 0; block < blocks; block++)
		michael_block(&mctx,
			data[block * 4] | data[block * 4 + 1] << 8 |
			data[block * 4 + 2] << 16 | data[block * 4 + 3] << 24);

	/* Partial block of 0..3 bytes and padding: 0x5a + 4..7 zeros to make
	 * total length a multiple of 4. */
	val = 0x5a;
	while (left > 0) {
		val <<= 8;
		left--;
		val |= data[blocks * 4 + left];
	}

	michael_block(&mctx, val);
	michael_block(&mctx, 0);

	mic[0] = mctx.l & 0xff;
	mic[1] = (mctx.l >> 8) & 0xff;
	mic[2] = (mctx.l >> 16) & 0xff;
	mic[3] = (mctx.l >> 24) & 0xff;
	mic[4] = mctx.r & 0xff;
	mic[5] = (mctx.r >> 8) & 0xff;
	mic[6] = (mctx.r >> 16) & 0xff;
	mic[7] = (mctx.r >> 24) & 0xff;
}
